@@ -34,32 +34,34 @@ if STDERR_FILE:
   logger.info('logging stderr in js compiler phase into %s' % STDERR_FILE)
   STDERR_FILE = open(STDERR_FILE, 'w')
 
+WASM_INIT_FUNC = '__wasm_call_ctors'
 
-def compute_minimal_runtime_initializer_and_exports(post, initializers, exports, receiving):
-  # Generate invocations for all global initializers directly off the asm export object, e.g. asm['__GLOBAL__INIT']();
-  post = post.replace('/*** RUN_GLOBAL_INITIALIZERS(); ***/', '\n'.join(["asm['" + x + "']();" for x in initializers]))
 
+def compute_minimal_runtime_initializer_and_exports(post, exports, receiving):
   # Declare all exports out to global JS scope so that JS library functions can access them in a
   # way that minifies well with Closure
   # e.g. var a,b,c,d,e,f;
-  exports_that_are_not_initializers = [x for x in exports if x not in initializers]
+  exports_that_are_not_initializers = [x for x in exports if x not in WASM_INIT_FUNC]
   # In Wasm backend the exports are still unmangled at this point, so mangle the names here
   exports_that_are_not_initializers = [asmjs_mangle(x) for x in exports_that_are_not_initializers]
-  post = post.replace('/*** ASM_MODULE_EXPORTS_DECLARES ***/', 'var ' + ',\n  '.join(exports_that_are_not_initializers) + ';')
+
+  # Decide whether we should generate the global dynCalls dictionary for the dynCall() function?
+  if shared.Settings.DYNCALLS and '$dynCall' in shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE and len([x for x in exports_that_are_not_initializers if x.startswith('dynCall_')]) > 0:
+    exports_that_are_not_initializers += ['dynCalls = {}']
+
+  declares = 'var ' + ',\n '.join(exports_that_are_not_initializers) + ';'
+  post = shared.do_replace(post, '<<< ASM_MODULE_EXPORTS_DECLARES >>>', declares)
 
   # Generate assignments from all asm.js/wasm exports out to the JS variables above: e.g. a = asm['a']; b = asm['b'];
-  post = post.replace('/*** ASM_MODULE_EXPORTS ***/', receiving)
+  post = shared.do_replace(post, '<<< ASM_MODULE_EXPORTS >>>', receiving)
   return post
 
 
-def write_output_file(outfile, post, module):
+def write_output_file(outfile, module):
   for i in range(len(module)): # do this loop carefully to save memory
     module[i] = normalize_line_endings(module[i])
     outfile.write(module[i])
 
-  post = normalize_line_endings(post)
-  outfile.write(post)
-
 
 def optimize_syscalls(declares, DEBUG):
   """Disables filesystem if only a limited subset of syscalls is used.
@@ -115,38 +117,25 @@ def update_settings_glue(metadata, DEBUG):
     shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE = []
 
   all_funcs = shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE + [shared.JS.to_nice_ident(d) for d in metadata['declares']]
-  implemented_funcs = [x[1:] for x in metadata['implementedFunctions']]
-  shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE = sorted(set(all_funcs).difference(implemented_funcs))
+  shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE = sorted(set(all_funcs).difference(metadata['exports']))
 
   shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += [x[1:] for x in metadata['externs']]
-  shared.Settings.IMPLEMENTED_FUNCTIONS = metadata['implementedFunctions']
+  # With the wasm backend the set of implemented functions is identical to the set of exports
+  shared.Settings.IMPLEMENTED_FUNCTIONS = [asmjs_mangle(x) for x in metadata['exports']]
 
   if metadata['asmConsts']:
     # emit the EM_ASM signature-reading helper function only if we have any EM_ASM
     # functions in the module.
     shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$readAsmConstArgs']
 
-    # Extract the list of function signatures that MAIN_THREAD_EM_ASM blocks in
-    # the compiled code have, each signature will need a proxy function invoker
-    # generated for it.
-    def read_proxied_function_signatures(asmConsts):
-      proxied_function_signatures = set()
-      for _, sigs, proxying_types in asmConsts.values():
-        for sig, proxying_type in zip(sigs, proxying_types):
-          if proxying_type == 'sync_on_main_thread_':
-            proxied_function_signatures.add(sig + '_sync')
-          elif proxying_type == 'async_on_main_thread_':
-            proxied_function_signatures.add(sig + '_async')
-      return list(proxied_function_signatures)
-
-    shared.Settings.PROXIED_FUNCTION_SIGNATURES = read_proxied_function_signatures(metadata['asmConsts'])
-
   shared.Settings.BINARYEN_FEATURES = metadata['features']
   if shared.Settings.RELOCATABLE:
     # When building relocatable output (e.g. MAIN_MODULE) the reported table
     # size does not include the reserved slot at zero for the null pointer.
     # Instead we use __table_base to offset the elements by 1.
-    shared.Settings.WASM_TABLE_SIZE = metadata['tableSize'] + 1
+    if shared.Settings.INITIAL_TABLE == -1:
+      shared.Settings.INITIAL_TABLE = metadata['tableSize'] + 1
+
   shared.Settings.MAIN_READS_PARAMS = metadata['mainReadsParams']
 
   # Store exports for Closure compiler to be able to track these as globals in
@@ -161,32 +150,17 @@ def update_settings_glue(metadata, DEBUG):
       shared.Settings.EXPORTED_RUNTIME_METHODS += ['writeStackCookie', 'checkStackCookie']
 
 
-# static code hooks
-class StaticCodeHooks:
-  atinits = []
-  atmains = []
-  atexits = []
-
-
-def apply_static_code_hooks(code):
-  code = code.replace('{{{ ATINITS }}}', StaticCodeHooks.atinits)
-  code = code.replace('{{{ ATMAINS }}}', StaticCodeHooks.atmains)
-  code = code.replace('{{{ ATEXITS }}}', StaticCodeHooks.atexits)
+def apply_static_code_hooks(forwarded_json, code):
+  code = shared.do_replace(code, '<<< ATINITS >>>', str(forwarded_json['ATINITS']))
+  code = shared.do_replace(code, '<<< ATMAINS >>>', str(forwarded_json['ATMAINS']))
+  if shared.Settings.EXIT_RUNTIME:
+    code = shared.do_replace(code, '<<< ATEXITS >>>', str(forwarded_json['ATEXITS']))
   return code
 
 
-def apply_forwarded_data(forwarded_data):
-  forwarded_json = json.loads(forwarded_data)
-  # Be aware of JS static allocations
-  # Be aware of JS static code hooks
-  StaticCodeHooks.atinits = str(forwarded_json['ATINITS'])
-  StaticCodeHooks.atmains = str(forwarded_json['ATMAINS'])
-  StaticCodeHooks.atexits = str(forwarded_json['ATEXITS'])
-
-
-def compile_settings(temp_files):
+def compile_settings():
   # Save settings to a file to work around v8 issue 1579
-  with temp_files.get_file('.txt') as settings_file:
+  with shared.configuration.get_temp_files().get_file('.txt') as settings_file:
     with open(settings_file, 'w') as s:
       json.dump(shared.Settings.to_dict(), s, sort_keys=True)
 
@@ -198,28 +172,15 @@ def compile_settings(temp_files):
                              cwd=path_from_root('src'), env=env)
   assert '//FORWARDED_DATA:' in out, 'Did not receive forwarded data in pre output - process failed?'
   glue, forwarded_data = out.split('//FORWARDED_DATA:')
-
-  apply_forwarded_data(forwarded_data)
-
   return glue, forwarded_data
 
 
-class Memory():
-  def __init__(self, static_bump):
-    stack_low = align_memory(shared.Settings.GLOBAL_BASE + static_bump)
-    stack_high = align_memory(stack_low + shared.Settings.TOTAL_STACK)
-    self.stack_base = stack_high
-    self.stack_max = stack_low
-    self.dynamic_base = align_memory(stack_high)
-
-
-def apply_memory(js, memory):
-  # Apply the statically-at-compile-time computed memory locations.
-  # Write it all out
-  js = js.replace('{{{ HEAP_BASE }}}', str(memory.dynamic_base))
-  js = js.replace('{{{ STACK_BASE }}}', str(memory.stack_base))
-  js = js.replace('{{{ STACK_MAX }}}', str(memory.stack_max))
-  return js
+def set_memory(static_bump):
+  stack_low = align_memory(shared.Settings.GLOBAL_BASE + static_bump)
+  stack_high = align_memory(stack_low + shared.Settings.TOTAL_STACK)
+  shared.Settings.STACK_BASE = stack_high
+  shared.Settings.STACK_MAX = stack_low
+  shared.Settings.HEAP_BASE = align_memory(stack_high)
 
 
 def report_missing_symbols(all_implemented, pre):
@@ -304,7 +265,7 @@ def create_named_globals(metadata):
   return '\n'.join(named_globals)
 
 
-def emscript(in_wasm, out_wasm, outfile_js, memfile, temp_files, DEBUG):
+def emscript(in_wasm, out_wasm, outfile_js, memfile, DEBUG):
   # Overview:
   #   * Run wasm-emscripten-finalize to extract metadata and modify the binary
   #     to use emscripten's wasm<->JS ABI
@@ -329,7 +290,14 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile, temp_files, DEBUG):
     logger.debug('emscript: js compiler glue')
     t = time.time()
 
-  glue, forwarded_data = compile_settings(temp_files)
+  # memory and global initializers
+
+  if shared.Settings.RELOCATABLE:
+    static_bump = align_memory(webassembly.parse_dylink_section(in_wasm)[0])
+    set_memory(static_bump)
+    logger.debug('stack_base: %d, stack_max: %d, heap_base: %d', shared.Settings.STACK_BASE, shared.Settings.STACK_MAX, shared.Settings.HEAP_BASE)
+
+  glue, forwarded_data = compile_settings()
   if DEBUG:
     logger.debug('  emscript: glue took %s seconds' % (time.time() - t))
     t = time.time()
@@ -355,28 +323,12 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile, temp_files, DEBUG):
     logger.debug('emscript: skipping remaining js glue generation')
     return
 
-  # memory and global initializers
-
-  # In minimal runtime, global initializers are run after the Wasm Module instantiation has finished.
-  if not shared.Settings.MINIMAL_RUNTIME:
-    global_initializers = ', '.join('{ func: function() { %s() } }' % i for i in metadata['initializers'])
-    # In regular runtime, global initializers are recorded in an __ATINIT__ array.
-    global_initializers = '__ATINIT__.push(%s);' % global_initializers
-    if shared.Settings.USE_PTHREADS:
-      global_initializers = 'if (!ENVIRONMENT_IS_PTHREAD) ' + global_initializers
-
-    pre += '\n' + global_initializers + '\n'
-
-  if shared.Settings.RELOCATABLE:
-    static_bump = align_memory(webassembly.parse_dylink_section(in_wasm)[0])
-    memory = Memory(static_bump)
-    logger.debug('stack_base: %d, stack_max: %d, dynamic_base: %d', memory.stack_base, memory.stack_max, memory.dynamic_base)
-
-    pre = apply_memory(pre, memory)
-    post = apply_memory(post, memory)
-
-  pre = apply_static_code_hooks(pre) # In regular runtime, atinits etc. exist in the preamble part
-  post = apply_static_code_hooks(post) # In MINIMAL_RUNTIME, atinit exists in the postamble part
+  if shared.Settings.MINIMAL_RUNTIME:
+    # In MINIMAL_RUNTIME, atinit exists in the postamble part
+    post = apply_static_code_hooks(forwarded_json, post)
+  else:
+    # In regular runtime, atinits etc. exist in the preamble part
+    pre = apply_static_code_hooks(forwarded_json, pre)
 
   # merge forwarded data
   shared.Settings.EXPORTED_FUNCTIONS = forwarded_json['EXPORTED_FUNCTIONS']
@@ -391,20 +343,23 @@ def emscript(in_wasm, out_wasm, outfile_js, memfile, temp_files, DEBUG):
      '\n'.join(em_js_funcs) + '\n'))
 
   with open(outfile_js, 'w') as out:
-    out.write(pre)
+    out.write(normalize_line_endings(pre))
     pre = None
 
     invoke_funcs = metadata['invokeFuncs']
     sending = create_sending(invoke_funcs, metadata)
-    receiving = create_receiving(exports, metadata['initializers'])
+    receiving = create_receiving(exports)
 
     if shared.Settings.MINIMAL_RUNTIME:
-      post = compute_minimal_runtime_initializer_and_exports(post, metadata['initializers'], exports, receiving)
+      if shared.Settings.DECLARE_ASM_MODULE_EXPORTS:
+        post = compute_minimal_runtime_initializer_and_exports(post, exports, receiving)
       receiving = ''
 
     module = create_module(sending, receiving, invoke_funcs, metadata)
 
-    write_output_file(out, post, module)
+    write_output_file(out, module)
+
+    out.write(normalize_line_endings(post))
     module = None
 
 
@@ -440,7 +395,7 @@ def finalize_wasm(infile, outfile, memfile, DEBUG):
     args.append('-g')
   if shared.Settings.WASM_BIGINT:
     args.append('--bigint')
-  if shared.Settings.USE_LEGACY_DYNCALLS:
+  if shared.Settings.DYNCALLS:
     # we need to add all dyncalls to the wasm
     modify_wasm = True
   else:
@@ -497,8 +452,7 @@ def finalize_wasm(infile, outfile, memfile, DEBUG):
 
 def create_asm_consts(metadata):
   asm_consts = {}
-  for k, v in metadata['asmConsts'].items():
-    const, sigs, call_types = v
+  for addr, const in metadata['asmConsts'].items():
     const = asstr(const)
     const = trim_asm_const_body(const)
     args = []
@@ -510,7 +464,7 @@ def create_asm_consts(metadata):
     for i in range(arity):
       args.append('$' + str(i))
     const = 'function(' + ', '.join(args) + ') {' + const + '}'
-    asm_consts[int(k)] = const
+    asm_consts[int(addr)] = const
   asm_consts = [(key, value) for key, value in asm_consts.items()]
   asm_consts.sort()
   return asm_consts
@@ -536,10 +490,7 @@ def create_em_js(forwarded_json, metadata):
 
 
 def add_standard_wasm_imports(send_items_map):
-  # Normally we import these into the wasm (so that JS could use them even
-  # before the wasm loads), while in standalone mode we do not depend
-  # on JS to create them, but create them in the wasm and export them.
-  if not shared.Settings.STANDALONE_WASM:
+  if shared.Settings.IMPORTED_MEMORY:
     memory_import = 'wasmMemory'
     if shared.Settings.MODULARIZE and shared.Settings.USE_PTHREADS:
       # Pthreads assign wasmMemory in their worker startup. In MODULARIZE mode, they cannot assign inside the
@@ -732,13 +683,13 @@ var %(mangled)s = Module["%(mangled)s"] = asm["%(name)s"]
   return wrappers
 
 
-def create_receiving(exports, initializers):
+def create_receiving(exports):
   # When not declaring asm exports this section is empty and we instead programatically export
   # symbols on the global object by calling exportAsmFunctions after initialization
   if not shared.Settings.DECLARE_ASM_MODULE_EXPORTS:
     return ''
 
-  exports_that_are_not_initializers = [x for x in exports if x not in initializers]
+  exports_that_are_not_initializers = [x for x in exports if x != WASM_INIT_FUNC]
 
   receiving = []
 
@@ -752,7 +703,11 @@ def create_receiving(exports, initializers):
       # WebAssembly.instantiate(Module["wasm"], imports).then((function(output) {
       # var asm = output.instance.exports;
       # _main = asm["_main"];
-      receiving += [asmjs_mangle(s) + ' = asm["' + s + '"];' for s in exports_that_are_not_initializers]
+      generate_dyncall_assignment = shared.Settings.DYNCALLS and '$dynCall' in shared.Settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE
+      for s in exports_that_are_not_initializers:
+        mangled = asmjs_mangle(s)
+        dynCallAssignment = ('dynCalls["' + s.replace('dynCall_', '') + '"] = ') if generate_dyncall_assignment and mangled.startswith('dynCall_') else ''
+        receiving += [dynCallAssignment + mangled + ' = asm["' + s + '"];']
     else:
       if shared.Settings.MINIMAL_RUNTIME:
         # In wasm2js exports can be directly processed at top level, i.e.
@@ -804,12 +759,9 @@ def load_metadata_wasm(metadata_raw, DEBUG):
 
   metadata = {
     'declares': [],
-    'implementedFunctions': [],
     'externs': [],
-    'simd': False, # Obsolete, always False
     'staticBump': 0,
     'tableSize': 0,
-    'initializers': [],
     'exports': [],
     'namedGlobals': {},
     'emJsFuncs': {},
@@ -818,9 +770,12 @@ def load_metadata_wasm(metadata_raw, DEBUG):
     'features': [],
     'mainReadsParams': 1,
   }
+  legacy_keys = set(['implementedFunctions', 'initializers', 'simd'])
 
   assert 'tableSize' in metadata_json.keys()
   for key, value in metadata_json.items():
+    if key in legacy_keys:
+      continue
     # json.loads returns `unicode` for strings but other code in this file
     # generally works with utf8 encoded `str` objects, and they don't alwasy
     # mix well.  e.g. s.replace(x, y) will blow up is `s` a uts8 str containing
@@ -833,10 +788,10 @@ def load_metadata_wasm(metadata_raw, DEBUG):
       exit_with_error('unexpected metadata key received from wasm-emscripten-finalize: %s', key)
     metadata[key] = value
 
-  if not shared.Settings.MINIMAL_RUNTIME:
-    # In regular runtime initializers call the global var version of the export, so they get the mangled name.
-    # In MINIMAL_RUNTIME, the initializers are called directly off the export object for minimal code size.
-    metadata['initializers'] = [asmjs_mangle(i) for i in metadata['initializers']]
+  # Support older metadata when asmConsts values were lists.  We only use the first element
+  # nowadays
+  # TODO(sbc): remove this once binaryen has been changed to only emit the single element
+  metadata['asmConsts'] = {k: v[0] if type(v) is list else v for k, v in metadata['asmConsts'].items()}
 
   if DEBUG:
     logger.debug("Metadata parsed: " + pprint.pformat(metadata))
@@ -849,10 +804,6 @@ def load_metadata_wasm(metadata_raw, DEBUG):
   unexpected_exports = [e for e in unexpected_exports if e not in shared.Settings.EXPORTED_FUNCTIONS]
   building.user_requested_exports += unexpected_exports
 
-  # With the wasm backend the set of implemented functions is identical to the set of exports
-  # Set this key here simply so that the shared code that handle it.
-  metadata['implementedFunctions'] = [asmjs_mangle(x) for x in metadata['exports']]
-
   return metadata
 
 
@@ -878,20 +829,15 @@ def normalize_line_endings(text):
 def generate_struct_info():
   generated_struct_info_name = 'generated_struct_info.json'
 
-  def generate_struct_info():
+  def generate_struct_info(out):
     with ToolchainProfiler.profile_block('gen_struct_info'):
-      out = shared.Cache.get_path(generated_struct_info_name)
-      gen_struct_info.main(['-q', '-c', '-o', out])
-      return out
+      gen_struct_info.main(['-q', '-o', out])
 
   shared.Settings.STRUCT_INFO = shared.Cache.get(generated_struct_info_name, generate_struct_info)
 
 
 def run(in_wasm, out_wasm, outfile_js, memfile):
-  temp_files = shared.configuration.get_temp_files()
   if not shared.Settings.BOOTSTRAPPING_STRUCT_INFO:
     generate_struct_info()
 
-  return temp_files.run_and_clean(lambda: emscript(
-      in_wasm, out_wasm, outfile_js, memfile, temp_files, shared.DEBUG)
-  )
+  emscript(in_wasm, out_wasm, outfile_js, memfile, shared.DEBUG)
