@@ -73,7 +73,7 @@ var LibraryGLEW = {
           string = "Unknown error";
           error = 8; // prevent array from growing more than this
         }
-        GLEW.error[error] = allocate(intArrayFromString(string), 'i8', ALLOC_NORMAL);
+        GLEW.error[error] = allocate(intArrayFromString(string), ALLOC_NORMAL);
       }
       return GLEW.error[error];
     },
@@ -93,7 +93,7 @@ var LibraryGLEW = {
         var string = GLEW.versionStringConstantFromCode(name);
         if (!string)
           return 0;
-        GLEW.version[name] = allocate(intArrayFromString(string), 'i8', ALLOC_NORMAL);
+        GLEW.version[name] = allocate(intArrayFromString(string), ALLOC_NORMAL);
       }
       return GLEW.version[name];
     },
@@ -127,6 +127,7 @@ var LibraryGLEW = {
     return GLEW.extensionIsSupported(UTF8ToString(name));
   },
 
+  glewGetErrorString__sig: 'ii',
   glewGetErrorString: function(error) {
     return GLEW.errorString(error);
   },
