@@ -6,7 +6,6 @@
 import glob
 import hashlib
 import itertools
-import json
 import logging
 import os
 import re
@@ -16,12 +15,11 @@ import tarfile
 import zipfile
 from glob import iglob
 
-from . import shared, building, ports, config
+from . import shared, building, ports, config, utils
+from . import deps_info, tempfiles
+from . import diagnostics
 from tools.shared import mangle_c_symbol_name, demangle_c_symbol_name
 
-stdout = None
-stderr = None
-
 logger = logging.getLogger('system_libs')
 
 LIBC_SOCKETS = ['socket.c', 'socketpair.c', 'shutdown.c', 'bind.c', 'connect.c',
@@ -38,7 +36,8 @@ def files_in_path(path_components, filenames):
 
 def glob_in_path(path_components, glob_pattern, excludes=()):
   srcdir = shared.path_from_root(*path_components)
-  return [f for f in iglob(os.path.join(srcdir, glob_pattern)) if os.path.basename(f) not in excludes]
+  files = iglob(os.path.join(srcdir, glob_pattern), recursive=True)
+  return [f for f in files if os.path.basename(f) not in excludes]
 
 
 def get_all_files_under(dirname):
@@ -61,6 +60,8 @@ def get_cflags(force_object_files=False):
     flags += ['-flto=' + shared.Settings.LTO]
   if shared.Settings.RELOCATABLE:
     flags += ['-s', 'RELOCATABLE']
+  if shared.Settings.MEMORY64:
+    flags += ['-s', 'MEMORY64=' + str(shared.Settings.MEMORY64)]
   return flags
 
 
@@ -76,12 +77,16 @@ def run_one_command(cmd):
       del safe_env[opt]
   # TODO(sbc): Remove this one we remove the test_em_config_env_var test
   cmd.append('-Wno-deprecated')
-  shared.run_process(cmd, stdout=stdout, stderr=stderr, env=safe_env)
+  shared.run_process(cmd, env=safe_env)
 
 
 def run_build_commands(commands):
+  # Before running a set of build commands make sure the common sysroot
+  # headers are installed.  This prevents each sub-process from attempting
+  # to setup the sysroot itself.
+  ensure_sysroot()
   cores = min(len(commands), building.get_num_cores())
-  if cores <= 1:
+  if cores <= 1 or shared.DEBUG:
     for command in commands:
       run_one_command(command)
   else:
@@ -166,11 +171,6 @@ def get_wasm_libc_rt_files():
   return math_files + other_files + iprintf_files
 
 
-def in_temp(*args):
-  """Gets the path of a file in our temporary directory."""
-  return os.path.join(shared.get_emscripten_temp_dir(), *args)
-
-
 class Library(object):
   """
   `Library` is the base class of all system libraries.
@@ -249,7 +249,11 @@ class Library(object):
 
   # A list of flags to pass to emcc.
   # The flags for the parent class is automatically inherited.
-  cflags = ['-Werror']
+  # TODO: Investigate whether perf gains from loop unrolling would be worth the
+  # extra code size. The -fno-unroll-loops flags was added here when loop
+  # unrolling landed upstream in LLVM to avoid changing behavior but was not
+  # specifically evaluated.
+  cflags = ['-Werror', '-fno-unroll-loops']
 
   # A list of directories to put in the include path when building.
   # This is a list of tuples of path components.
@@ -292,20 +296,6 @@ class Library(object):
     if not self.name:
       raise NotImplementedError('Cannot instantiate an abstract library')
 
-    # Read .symbols file if it exists. This first tries to read a symbols file
-    # with the same basename with the library file name (e.g.
-    # libc++-mt.symbols), and if there isn't one, it tries to read the 'default'
-    # symbol file, which does not have any optional suffices (e.g.
-    # libc++.symbols).
-    basename = shared.unsuffixed(self.get_filename())
-    symbols_dir = shared.path_from_root('system', 'lib', 'symbols', 'wasm')
-    symbols_file = os.path.join(symbols_dir, basename + '.symbols')
-    default_symbols_file = os.path.join(symbols_dir, self.name + '.symbols')
-    if os.path.isfile(symbols_file):
-      self.symbols = read_symbols(symbols_file)
-    elif os.path.isfile(default_symbols_file):
-      self.symbols = read_symbols(default_symbols_file)
-
   def can_use(self):
     """
     Whether this library can be used in the current environment.
@@ -324,7 +314,7 @@ class Library(object):
     return True
 
   def erase(self):
-    shared.Cache.erase_file(self.get_filename())
+    shared.Cache.erase_file(shared.Cache.get_lib_name(self.get_filename()))
 
   def get_path(self):
     """
@@ -332,7 +322,7 @@ class Library(object):
 
     This will trigger a build if this library is not in the cache.
     """
-    return shared.Cache.get(self.get_filename(), self.build)
+    return shared.Cache.get_lib(self.get_filename(), self.build)
 
   def get_files(self):
     """
@@ -352,7 +342,7 @@ class Library(object):
 
     raise NotImplementedError()
 
-  def build_objects(self):
+  def build_objects(self, build_dir):
     """
     Returns a list of compiled object files for this library.
 
@@ -363,7 +353,7 @@ class Library(object):
     objects = []
     cflags = self.get_cflags()
     for src in self.get_files():
-      o = in_temp(shared.unsuffixed_basename(src) + '.o')
+      o = os.path.join(build_dir, shared.unsuffixed_basename(src) + '.o')
       ext = shared.suffix(src)
       if ext in ('.s', '.c'):
         cmd = [shared.EMCC]
@@ -371,16 +361,20 @@ class Library(object):
         cmd = [shared.EMXX]
       if ext != '.s':
         cmd += cflags
+      elif shared.Settings.MEMORY64:
+        cmd += ['-s', 'MEMORY64=' + str(shared.Settings.MEMORY64)]
       commands.append(cmd + ['-c', src, '-o', o])
       objects.append(o)
     run_build_commands(commands)
     return objects
 
-  def build(self):
+  def build(self, out_filename):
     """Builds the library and returns the path to the file."""
-    out_filename = in_temp(self.get_filename())
-    create_lib(out_filename, self.build_objects())
-    return out_filename
+    build_dir = shared.Cache.get_path(os.path.join('build', self.get_base_name()))
+    utils.safe_ensure_dirs(build_dir)
+    create_lib(out_filename, self.build_objects(build_dir))
+    if not shared.DEBUG:
+      tempfiles.try_delete(build_dir)
 
   @classmethod
   def _inherit_list(cls, attr):
@@ -519,7 +513,7 @@ class MTLibrary(Library):
   def get_cflags(self):
     cflags = super(MTLibrary, self).get_cflags()
     if self.is_mt:
-      cflags += ['-s', 'USE_PTHREADS=1', '-DUSE_THREADS']
+      cflags += ['-s', 'USE_PTHREADS']
     return cflags
 
   def get_base_name(self):
@@ -583,9 +577,9 @@ class NoExceptLibrary(Library):
   @classmethod
   def variations(cls, **kwargs):
     combos = super(NoExceptLibrary, cls).variations()
+    # TODO Reenable 'wasm' option after the new EH implementation is stabilized
     return ([dict(eh_mode=exceptions.none, **combo) for combo in combos] +
-            [dict(eh_mode=exceptions.emscripten, **combo) for combo in combos] +
-            [dict(eh_mode=exceptions.wasm, **combo) for combo in combos])
+            [dict(eh_mode=exceptions.emscripten, **combo) for combo in combos])
 
   @classmethod
   def get_default_variation(cls, **kwargs):
@@ -644,7 +638,7 @@ class libcompiler_rt(MTLibrary):
   cflags = ['-O2', '-fno-builtin']
   src_dir = ['system', 'lib', 'compiler-rt', 'lib', 'builtins']
   src_files = glob_in_path(src_dir, '*.c')
-  src_files.append(shared.path_from_root('system', 'lib', 'compiler-rt', 'stack_ops.s'))
+  src_files.append(shared.path_from_root('system', 'lib', 'compiler-rt', 'stack_ops.S'))
   src_files.append(shared.path_from_root('system', 'lib', 'compiler-rt', 'stack_limits.S'))
   src_files.append(shared.path_from_root('system', 'lib', 'compiler-rt', 'emscripten_setjmp.c'))
   src_files.append(shared.path_from_root('system', 'lib', 'compiler-rt', 'emscripten_exception_builtins.c'))
@@ -684,7 +678,7 @@ class libc(AsanInstrumentedLibrary, MuslInternalLibrary, MTLibrary):
         'inet_addr.c', 'res_query.c', 'res_querydomain.c', 'gai_strerror.c',
         'proto.c', 'gethostbyaddr.c', 'gethostbyaddr_r.c', 'gethostbyname.c',
         'gethostbyname2_r.c', 'gethostbyname_r.c', 'gethostbyname2.c',
-        'usleep.c', 'alarm.c', 'syscall.c', '_exit.c', 'popen.c',
+        'alarm.c', 'syscall.c', '_exit.c', 'popen.c',
         'getgrouplist.c', 'initgroups.c', 'wordexp.c', 'timer_create.c',
         'faccessat.c',
         # 'process' exclusion
@@ -730,7 +724,7 @@ class libc(AsanInstrumentedLibrary, MuslInternalLibrary, MTLibrary):
     # Allowed files from ignored modules
     libc_files += files_in_path(
         path_components=['system', 'lib', 'libc', 'musl', 'src', 'time'],
-        filenames=['clock_settime.c', 'asctime.c', 'ctime.c', 'gmtime.c', 'localtime.c'])
+        filenames=['clock_settime.c', 'asctime.c', 'ctime.c', 'gmtime.c', 'localtime.c', 'nanosleep.c'])
     libc_files += files_in_path(
         path_components=['system', 'lib', 'libc', 'musl', 'src', 'legacy'],
         filenames=['getpagesize.c', 'err.c'])
@@ -745,7 +739,77 @@ class libc(AsanInstrumentedLibrary, MuslInternalLibrary, MTLibrary):
 
     libc_files += files_in_path(
         path_components=['system', 'lib', 'libc'],
-        filenames=['extras.c', 'wasi-helpers.c'])
+        filenames=['extras.c', 'wasi-helpers.c', 'emscripten_pthread.c'])
+
+    libc_files += files_in_path(
+        path_components=['system', 'lib', 'pthread'],
+        filenames=['emscripten_atomic.c'])
+
+    libc_files += files_in_path(
+        path_components=['system', 'lib', 'libc', 'musl', 'src', 'thread'],
+        filenames=[
+          'pthread_self.c',
+          # C11 thread library functions
+          'thrd_create.c',
+          'thrd_exit.c',
+          'thrd_join.c',
+          'thrd_sleep.c',
+          'thrd_yield.c',
+          'call_once.c',
+        ])
+
+    libc_files += glob_in_path(['system', 'lib', 'libc', 'compat'], '*.c')
+
+    if self.is_mt:
+      libc_files += files_in_path(
+        path_components=['system', 'lib', 'libc', 'musl', 'src', 'thread'],
+        filenames=[
+          'pthread_attr_destroy.c', 'pthread_condattr_setpshared.c',
+          'pthread_mutex_lock.c', 'pthread_spin_destroy.c', 'pthread_attr_get.c',
+          'pthread_cond_broadcast.c', 'pthread_mutex_setprioceiling.c',
+          'pthread_spin_init.c', 'pthread_attr_init.c', 'pthread_cond_destroy.c',
+          'pthread_mutex_timedlock.c', 'pthread_spin_lock.c',
+          'pthread_attr_setdetachstate.c', 'pthread_cond_init.c',
+          'pthread_mutex_trylock.c', 'pthread_spin_trylock.c',
+          'pthread_attr_setguardsize.c', 'pthread_cond_signal.c',
+          'pthread_mutex_unlock.c', 'pthread_spin_unlock.c',
+          'pthread_attr_setinheritsched.c', 'pthread_cond_timedwait.c',
+          'pthread_once.c', 'sem_destroy.c', 'pthread_attr_setschedparam.c',
+          'pthread_cond_wait.c', 'pthread_rwlockattr_destroy.c', 'sem_getvalue.c',
+          'pthread_attr_setschedpolicy.c', 'pthread_equal.c', 'pthread_rwlockattr_init.c',
+          'sem_init.c', 'pthread_attr_setscope.c', 'pthread_getspecific.c',
+          'pthread_rwlockattr_setpshared.c', 'sem_open.c', 'pthread_attr_setstack.c',
+          'pthread_key_create.c', 'pthread_rwlock_destroy.c', 'sem_post.c',
+          'pthread_attr_setstacksize.c', 'pthread_mutexattr_destroy.c',
+          'pthread_rwlock_init.c', 'sem_timedwait.c', 'pthread_barrierattr_destroy.c',
+          'pthread_mutexattr_init.c', 'pthread_rwlock_rdlock.c', 'sem_trywait.c',
+          'pthread_barrierattr_init.c', 'pthread_mutexattr_setprotocol.c',
+          'pthread_rwlock_timedrdlock.c', 'sem_unlink.c',
+          'pthread_barrierattr_setpshared.c', 'pthread_mutexattr_setpshared.c',
+          'pthread_rwlock_timedwrlock.c', 'sem_wait.c', 'pthread_barrier_destroy.c',
+          'pthread_mutexattr_setrobust.c', 'pthread_rwlock_tryrdlock.c',
+          '__timedwait.c', 'pthread_barrier_init.c', 'pthread_mutexattr_settype.c',
+          'pthread_rwlock_trywrlock.c', 'vmlock.c', 'pthread_barrier_wait.c',
+          'pthread_mutex_consistent.c', 'pthread_rwlock_unlock.c', '__wait.c',
+          'pthread_condattr_destroy.c', 'pthread_mutex_destroy.c',
+          'pthread_rwlock_wrlock.c', 'pthread_condattr_init.c',
+          'pthread_mutex_getprioceiling.c', 'pthread_setcanceltype.c',
+          'pthread_condattr_setclock.c', 'pthread_mutex_init.c',
+          'pthread_setspecific.c', 'pthread_setcancelstate.c',
+          'pthread_getconcurrency.c', 'pthread_setconcurrency.c',
+          'pthread_getschedparam.c', 'pthread_setschedparam.c',
+          'pthread_setschedprio.c', 'pthread_atfork.c',
+          'pthread_getcpuclockid.c',
+        ])
+      libc_files += files_in_path(
+        path_components=['system', 'lib', 'pthread'],
+        filenames=[
+          'library_pthread.c',
+          'emscripten_tls_init.c',
+          'emscripten_thread_state.s',
+        ])
+    else:
+      libc_files += [shared.path_from_root('system', 'lib', 'pthread', 'library_pthread_stub.c')]
 
     return libc_files
 
@@ -870,51 +934,18 @@ class libcxx(NoExceptLibrary, MTLibrary):
   cflags = ['-DLIBCXX_BUILDING_LIBCXXABI=1', '-D_LIBCPP_BUILDING_LIBRARY', '-Oz',
             '-D_LIBCPP_DISABLE_VISIBILITY_ANNOTATIONS']
 
-  src_dir = ['system', 'lib', 'libcxx']
-  src_files = [
-    'algorithm.cpp',
-    'any.cpp',
-    'bind.cpp',
-    'charconv.cpp',
-    'chrono.cpp',
-    'condition_variable.cpp',
-    'condition_variable_destructor.cpp',
-    'debug.cpp',
-    'exception.cpp',
-    'functional.cpp',
-    'future.cpp',
-    'hash.cpp',
-    'ios.cpp',
-    'iostream.cpp',
-    'locale.cpp',
-    'memory.cpp',
-    'mutex.cpp',
-    'mutex_destructor.cpp',
-    'new.cpp',
-    'optional.cpp',
-    'random.cpp',
-    'regex.cpp',
-    'shared_mutex.cpp',
-    'stdexcept.cpp',
-    'string.cpp',
-    'strstream.cpp',
-    'system_error.cpp',
-    'thread.cpp',
-    'typeinfo.cpp',
-    'utility.cpp',
-    'valarray.cpp',
-    'variant.cpp',
-    'vector.cpp',
-    os.path.join('experimental', 'memory_resource.cpp'),
-    os.path.join('filesystem', 'directory_iterator.cpp'),
-    os.path.join('filesystem', 'operations.cpp')
-  ]
+  src_dir = ['system', 'lib', 'libcxx', 'src']
+  src_glob = '**/*.cpp'
+  src_glob_exclude = ['locale_win32.cpp', 'thread_win32.cpp', 'support.cpp', 'int128_builtins.cpp']
 
 
 class libunwind(NoExceptLibrary, MTLibrary):
   name = 'libunwind'
   cflags = ['-Oz', '-D_LIBUNWIND_DISABLE_VISIBILITY_ANNOTATIONS']
   src_dir = ['system', 'lib', 'libunwind', 'src']
+  # Without this we can't build libunwind since it will pickup the unwind.h
+  # that is part of llvm (which is not compatible for some reason).
+  includes = [['system', 'lib', 'libunwind', 'include']]
   src_files = ['Unwind-wasm.cpp']
 
   def __init__(self, **kwargs):
@@ -944,36 +975,39 @@ class libmalloc(MTLibrary):
 
   def __init__(self, **kwargs):
     self.malloc = kwargs.pop('malloc')
-    if self.malloc not in ('dlmalloc', 'emmalloc', 'none'):
-      raise Exception('malloc must be one of "emmalloc", "dlmalloc" or "none", see settings.js')
+    if self.malloc not in ('dlmalloc', 'emmalloc', 'emmalloc-debug', 'emmalloc-memvalidate', 'emmalloc-verbose', 'emmalloc-memvalidate-verbose', 'none'):
+      raise Exception('malloc must be one of "emmalloc[-debug|-memvalidate][-verbose]", "dlmalloc" or "none", see settings.js')
 
-    self.is_debug = kwargs.pop('is_debug')
     self.use_errno = kwargs.pop('use_errno')
     self.is_tracing = kwargs.pop('is_tracing')
-    self.use_64bit_ops = kwargs.pop('use_64bit_ops')
+    self.memvalidate = kwargs.pop('memvalidate')
+    self.verbose = kwargs.pop('verbose')
+    self.is_debug = kwargs.pop('is_debug') or self.memvalidate or self.verbose
 
     super(libmalloc, self).__init__(**kwargs)
 
   def get_files(self):
+    malloc_base = self.malloc.replace('-memvalidate', '').replace('-verbose', '').replace('-debug', '')
     malloc = shared.path_from_root('system', 'lib', {
-      'dlmalloc': 'dlmalloc.c', 'emmalloc': 'emmalloc.cpp'
-    }[self.malloc])
+      'dlmalloc': 'dlmalloc.c', 'emmalloc': 'emmalloc.cpp',
+    }[malloc_base])
     sbrk = shared.path_from_root('system', 'lib', 'sbrk.c')
     return [malloc, sbrk]
 
   def get_cflags(self):
     cflags = super(libmalloc, self).get_cflags()
+    if self.memvalidate:
+      cflags += ['-DEMMALLOC_MEMVALIDATE']
+    if self.verbose:
+      cflags += ['-DEMMALLOC_VERBOSE']
     if self.is_debug:
       cflags += ['-UNDEBUG', '-DDLMALLOC_DEBUG']
-      # TODO: consider adding -DEMMALLOC_DEBUG, but that is quite slow
     else:
       cflags += ['-DNDEBUG']
     if not self.use_errno:
       cflags += ['-DMALLOC_FAILURE_ACTION=', '-DEMSCRIPTEN_NO_ERRNO']
     if self.is_tracing:
       cflags += ['--tracing']
-    if self.use_64bit_ops:
-      cflags += ['-DEMMALLOC_USE_64BIT_OPS=1']
     return cflags
 
   def get_base_name_prefix(self):
@@ -981,15 +1015,13 @@ class libmalloc(MTLibrary):
 
   def get_base_name(self):
     name = super(libmalloc, self).get_base_name()
-    if self.is_debug:
+    if self.is_debug and not self.memvalidate and not self.verbose:
       name += '-debug'
     if not self.use_errno:
       # emmalloc doesn't actually use errno, but it's easier to build it again
       name += '-noerrno'
     if self.is_tracing:
       name += '-tracing'
-    if self.use_64bit_ops:
-      name += '-64bit'
     return name
 
   def can_use(self):
@@ -997,7 +1029,7 @@ class libmalloc(MTLibrary):
 
   @classmethod
   def vary_on(cls):
-    return super(libmalloc, cls).vary_on() + ['is_debug', 'use_errno', 'is_tracing', 'use_64bit_ops']
+    return super(libmalloc, cls).vary_on() + ['is_debug', 'use_errno', 'is_tracing', 'memvalidate', 'verbose']
 
   @classmethod
   def get_default_variation(cls, **kwargs):
@@ -1006,15 +1038,19 @@ class libmalloc(MTLibrary):
       is_debug=shared.Settings.ASSERTIONS >= 2,
       use_errno=shared.Settings.SUPPORT_ERRNO,
       is_tracing=shared.Settings.EMSCRIPTEN_TRACING,
-      use_64bit_ops=shared.Settings.MALLOC == 'emmalloc' and (shared.Settings.WASM == 1 or shared.Settings.WASM2JS == 0),
+      memvalidate='memvalidate' in shared.Settings.MALLOC,
+      verbose='verbose' in shared.Settings.MALLOC,
       **kwargs
     )
 
   @classmethod
   def variations(cls):
     combos = super(libmalloc, cls).variations()
-    return ([dict(malloc='dlmalloc', **combo) for combo in combos if not combo['use_64bit_ops']] +
-            [dict(malloc='emmalloc', **combo) for combo in combos])
+    return ([dict(malloc='dlmalloc', **combo) for combo in combos if not combo['memvalidate'] and not combo['verbose']] +
+            [dict(malloc='emmalloc', **combo) for combo in combos if not combo['memvalidate'] and not combo['verbose']] +
+            [dict(malloc='emmalloc-memvalidate-verbose', **combo) for combo in combos if combo['memvalidate'] and combo['verbose']] +
+            [dict(malloc='emmalloc-memvalidate', **combo) for combo in combos if combo['memvalidate'] and not combo['verbose']] +
+            [dict(malloc='emmalloc-verbose', **combo) for combo in combos if combo['verbose'] and not combo['memvalidate']])
 
 
 class libal(Library):
@@ -1029,7 +1065,7 @@ class libgl(MTLibrary):
   name = 'libgl'
 
   src_dir = ['system', 'lib', 'gl']
-  src_glob = '*.c'
+  src_files = ['gl.c', 'webgl1.c']
 
   cflags = ['-Oz']
 
@@ -1038,6 +1074,10 @@ class libgl(MTLibrary):
     self.is_webgl2 = kwargs.pop('is_webgl2')
     self.is_ofb = kwargs.pop('is_ofb')
     self.is_full_es3 = kwargs.pop('is_full_es3')
+    if self.is_webgl2 or self.is_full_es3:
+      # Don't use append or += here, otherwise we end up adding to
+      # the class member.
+      self.src_files = self.src_files + ['webgl2.c']
     super(libgl, self).__init__(**kwargs)
 
   def get_base_name(self):
@@ -1148,58 +1188,6 @@ class libhtml5(Library):
   src_glob = '*.c'
 
 
-class libpthread(AsanInstrumentedLibrary, MuslInternalLibrary, MTLibrary):
-  name = 'libpthread'
-  cflags = ['-O2']
-
-  def get_files(self):
-    if self.is_mt:
-      files = files_in_path(
-        path_components=['system', 'lib', 'libc', 'musl', 'src', 'thread'],
-        filenames=[
-          'pthread_attr_destroy.c', 'pthread_condattr_setpshared.c',
-          'pthread_mutex_lock.c', 'pthread_spin_destroy.c', 'pthread_attr_get.c',
-          'pthread_cond_broadcast.c', 'pthread_mutex_setprioceiling.c',
-          'pthread_spin_init.c', 'pthread_attr_init.c', 'pthread_cond_destroy.c',
-          'pthread_mutex_timedlock.c', 'pthread_spin_lock.c',
-          'pthread_attr_setdetachstate.c', 'pthread_cond_init.c',
-          'pthread_mutex_trylock.c', 'pthread_spin_trylock.c',
-          'pthread_attr_setguardsize.c', 'pthread_cond_signal.c',
-          'pthread_mutex_unlock.c', 'pthread_spin_unlock.c',
-          'pthread_attr_setinheritsched.c', 'pthread_cond_timedwait.c',
-          'pthread_once.c', 'sem_destroy.c', 'pthread_attr_setschedparam.c',
-          'pthread_cond_wait.c', 'pthread_rwlockattr_destroy.c', 'sem_getvalue.c',
-          'pthread_attr_setschedpolicy.c', 'pthread_equal.c', 'pthread_rwlockattr_init.c',
-          'sem_init.c', 'pthread_attr_setscope.c', 'pthread_getspecific.c',
-          'pthread_rwlockattr_setpshared.c', 'sem_open.c', 'pthread_attr_setstack.c',
-          'pthread_key_create.c', 'pthread_rwlock_destroy.c', 'sem_post.c',
-          'pthread_attr_setstacksize.c', 'pthread_mutexattr_destroy.c',
-          'pthread_rwlock_init.c', 'sem_timedwait.c', 'pthread_barrierattr_destroy.c',
-          'pthread_mutexattr_init.c', 'pthread_rwlock_rdlock.c', 'sem_trywait.c',
-          'pthread_barrierattr_init.c', 'pthread_mutexattr_setprotocol.c',
-          'pthread_rwlock_timedrdlock.c', 'sem_unlink.c',
-          'pthread_barrierattr_setpshared.c', 'pthread_mutexattr_setpshared.c',
-          'pthread_rwlock_timedwrlock.c', 'sem_wait.c', 'pthread_barrier_destroy.c',
-          'pthread_mutexattr_setrobust.c', 'pthread_rwlock_tryrdlock.c',
-          '__timedwait.c', 'pthread_barrier_init.c', 'pthread_mutexattr_settype.c',
-          'pthread_rwlock_trywrlock.c', 'vmlock.c', 'pthread_barrier_wait.c',
-          'pthread_mutex_consistent.c', 'pthread_rwlock_unlock.c', '__wait.c',
-          'pthread_condattr_destroy.c', 'pthread_mutex_destroy.c',
-          'pthread_rwlock_wrlock.c', 'pthread_condattr_init.c',
-          'pthread_mutex_getprioceiling.c', 'pthread_setcanceltype.c',
-          'pthread_condattr_setclock.c', 'pthread_mutex_init.c',
-          'pthread_setspecific.c', 'pthread_setcancelstate.c'
-        ])
-      files += [shared.path_from_root('system', 'lib', 'pthread', 'library_pthread.c')]
-      files += [shared.path_from_root('system', 'lib', 'pthread', 'library_pthread_wasm.c')]
-      return files
-    else:
-      return [shared.path_from_root('system', 'lib', 'pthread', 'library_pthread_stub.c')]
-
-  def get_base_name_prefix(self):
-    return 'libpthread' if self.is_mt else 'libpthread_stub'
-
-
 class CompilerRTLibrary(Library):
   cflags = ['-O2', '-fno-builtin']
   # compiler_rt files can't currently be part of LTO although we are hoping to remove this
@@ -1341,7 +1329,6 @@ class libstandalonewasm(MuslInternalLibrary):
                    'gettimeofday.c',
                    'gmtime_r.c',
                    'localtime_r.c',
-                   'nanosleep.c',
                    'mktime.c',
                    'time.c'])
     # It is more efficient to use JS for __assert_fail, as it avoids always
@@ -1377,69 +1364,25 @@ def warn_on_unexported_main(symbolses):
         return
 
 
-def calculate(temp_files, cxx, forced, stdout_=None, stderr_=None):
-  global stdout, stderr
-  stdout = stdout_
-  stderr = stderr_
+def handle_reverse_deps(input_files):
+  if shared.Settings.REVERSE_DEPS == 'none':
+    return
+  elif shared.Settings.REVERSE_DEPS == 'all':
+    # When not optimzing we add all possible reverse dependencies rather
+    # than scanning the input files
+    for symbols in deps_info.deps_info.values():
+      for symbol in symbols:
+        shared.Settings.EXPORTED_FUNCTIONS.append(mangle_c_symbol_name(symbol))
+    return
 
-  # Setting this will only use the forced libs in EMCC_FORCE_STDLIBS. This avoids spending time checking
-  # for unresolved symbols in your project files, which can speed up linking, but if you do not have
-  # the proper list of actually needed libraries, errors can occur. See below for how we must
-  # export all the symbols in deps_info when using this option.
-  only_forced = os.environ.get('EMCC_ONLY_FORCED_STDLIBS')
-  if only_forced:
-    temp_files = []
+  if shared.Settings.REVERSE_DEPS != 'auto':
+    shared.exit_with_error(f'invalid values for REVERSE_DEPS: {shared.Settings.REVERSE_DEPS}')
 
-  # deps_info.json is a mechanism that lets JS code depend on C functions. This
-  # needs special help because of how linking works:
-  #
-  #   1. Receive some input files (.o, .c, etc.) from the user.
-  #   2. Link them with system libraries.
-  #   3. Whatever symbols are still unresolved, look in JS libraries for them.
-  #
-  # This makes C->JS calls work in a natural way: if compiled code depends on
-  # a function foo() that is implemented in a JS library, it will be unresolved
-  # after stage 2, and therefore linked in at stage 3. The problem is the other
-  # direction: if a JS library function decides it needs some function from say
-  # libc, then at stage 3 it is too late to link in more libc code. That's
-  # where deps_info.json comes in.
-  #
-  # Specifically, before stage 2 (linking with system libraries) we look at what
-  # symbols are required by the input files. Imagine that js_func in a JS
-  # library depends on libc_func in libc. Then if deps_info.json tells us
-  #
-  #  "js_func": ["libc_func"]
-  #
-  # then if we see js_func is required (undefined) before stage 2, then we add
-  # a requirement to link in libc_func when linking in system libraries. All
-  # we do with deps_info.json is see if any of the keys are among the
-  # undefined symbols before stage 2, and if they are, add their values to the
-  # things we need to link in.
-  #
-  # This usually works the way you want, but note that it happens *before* stage
-  # 2 and not after it. That is, we look for js_func before linking in system
-  # libraries. If you have a situation where
-  #
-  #   user_code => other_libc_func => js_func => libc_func
-  #
-  # then the deps_info.json entry must contain
-  #
-  #  "other_libc_func": ["libc_func"]
-  #
-  # because that is what we see before stage 2: all we see is that
-  # other_libc_func is going to be linked in, and we don't know yet that it
-  # will end up calling js_func. But the presence of a call to other_libc_func
-  # indicates that we will need libc_func linked in as well, so that is what the
-  # deps_info.json entry should contain.
-  #
-  # TODO: Move all __deps from src/library*.js to deps_info.json, and use that single source of info
-  #       both here and in the JS compiler.
-  deps_info = json.loads(open(shared.path_from_root('src', 'deps_info.json')).read())
   added = set()
 
-  def add_back_deps(need):
+  def add_reverse_deps(need):
     more = False
-    for ident, deps in deps_info.items():
+    for ident, deps in deps_info.deps_info.items():
       if ident in need.undefs and ident not in added:
         added.add(ident)
         more = True
@@ -1448,10 +1391,10 @@ def calculate(temp_files, cxx, forced, stdout_=None, stderr_=None):
           logger.debug('adding dependency on %s due to deps-info on %s' % (dep, ident))
           shared.Settings.EXPORTED_FUNCTIONS.append(mangle_c_symbol_name(dep))
     if more:
-      add_back_deps(need) # recurse to get deps of deps
+      add_reverse_deps(need) # recurse to get deps of deps
 
   # Scan symbols
-  symbolses = building.parallel_llvm_nm([os.path.abspath(t) for t in temp_files])
+  symbolses = building.llvm_nm_multiple([os.path.abspath(t) for t in input_files])
 
   warn_on_unexported_main(symbolses)
 
@@ -1468,15 +1411,22 @@ def calculate(temp_files, cxx, forced, stdout_=None, stderr_=None):
     symbolses[0].undefs.add(demangle_c_symbol_name(export))
 
   for symbols in symbolses:
-    add_back_deps(symbols)
+    add_reverse_deps(symbols)
 
-  # If we are only doing forced stdlibs, then we don't know the actual symbols we need,
-  # and must assume all of deps_info must be exported. Note that this might cause
-  # warnings on exports that do not exist.
+
+def calculate(input_files, cxx, forced):
+  # Setting this will only use the forced libs in EMCC_FORCE_STDLIBS. This avoids spending time checking
+  # for unresolved symbols in your project files, which can speed up linking, but if you do not have
+  # the proper list of actually needed libraries, errors can occur. See below for how we must
+  # export all the symbols in deps_info when using this option.
+  only_forced = os.environ.get('EMCC_ONLY_FORCED_STDLIBS')
   if only_forced:
-    for key, value in deps_info.items():
-      for dep in value:
-        shared.Settings.EXPORTED_FUNCTIONS.append(mangle_c_symbol_name(dep))
+    # One of the purposes EMCC_ONLY_FORCED_STDLIBS was to skip the scanning
+    # of the input files for reverse dependencies.
+    diagnostics.warning('deprecated', 'EMCC_ONLY_FORCED_STDLIBS is deprecated.  Use `-nostdlib` and/or `-s REVERSE_DEPS=none` depending on the desired result')
+    shared.Settings.REVERSE_DEPS = 'all'
+
+  handle_reverse_deps(input_files)
 
   libs_to_link = []
   already_included = set()
@@ -1518,39 +1468,13 @@ def calculate(temp_files, cxx, forced, stdout_=None, stderr_=None):
     add_library(system_libs_map[forced])
 
   if only_forced:
-    if not shared.Settings.BOOTSTRAPPING_STRUCT_INFO:
-      add_library(system_libs_map['libc_rt_wasm'])
+    add_library(system_libs_map['libc_rt_wasm'])
     add_library(system_libs_map['libcompiler_rt'])
   else:
-    # These libraries get included automatically based the symbols needed by the input file.
-    # Other system libraries are simply included by default.
-    for libname in ['libgl', 'libal', 'libhtml5']:
-      if libname in already_included:
-        continue
-      lib = system_libs_map[libname]
-      force_this = lib.name in force_include
-      if not force_this:
-        need_syms = set()
-        has_syms = set()
-        for symbols in symbolses:
-          if shared.Settings.VERBOSE:
-            logger.debug('undefs: ' + str(symbols.undefs))
-          for library_symbol in lib.symbols:
-            if library_symbol in symbols.undefs:
-              need_syms.add(library_symbol)
-            if library_symbol in symbols.defs:
-              has_syms.add(library_symbol)
-        for haz in has_syms:
-          if haz in need_syms:
-            # remove symbols that are supplied by another of the inputs
-            need_syms.remove(haz)
-        if shared.Settings.VERBOSE:
-          logger.debug('considering %s: we need %s and have %s' % (lib.name, str(need_syms), str(has_syms)))
-        if not len(need_syms):
-          continue
-
-      # We need to build and link the library in
-      add_library(lib)
+    if shared.Settings.AUTO_NATIVE_LIBRARIES:
+      add_library(system_libs_map['libgl'])
+      add_library(system_libs_map['libal'])
+      add_library(system_libs_map['libhtml5'])
 
     sanitize = shared.Settings.USE_LSAN or shared.Settings.USE_ASAN or shared.Settings.UBSAN_RUNTIME
 
@@ -1573,7 +1497,6 @@ def calculate(temp_files, cxx, forced, stdout_=None, stderr_=None):
         add_library(system_libs_map['libunwind'])
     if shared.Settings.MALLOC != 'none':
       add_library(system_libs_map['libmalloc'])
-    add_library(system_libs_map['libpthread'])
     if shared.Settings.STANDALONE_WASM:
       add_library(system_libs_map['libstandalonewasm'])
     add_library(system_libs_map['libc_rt_wasm'])
@@ -1642,13 +1565,9 @@ class Ports(object):
   """emscripten-ports library management (https://github.com/emscripten-ports).
   """
 
-  @staticmethod
-  def get_lib_name(name):
-    return name + '.a'
-
   @staticmethod
   def get_include_dir():
-    dirname = shared.Cache.get_path('include')
+    dirname = shared.Cache.get_include_dir()
     shared.safe_ensure_dirs(dirname)
     return dirname
 
@@ -1753,10 +1672,9 @@ class Ports(object):
     # clears the build, so that it is rebuilt from that source.
     local_ports = os.environ.get('EMCC_LOCAL_PORTS')
     if local_ports:
-      shared.Cache.acquire_cache_lock()
       logger.warning('using local ports: %s' % local_ports)
       local_ports = [pair.split('=', 1) for pair in local_ports.split(',')]
-      try:
+      with shared.Cache.lock():
         for local in local_ports:
           if name == local[0]:
             path = local[1]
@@ -1776,8 +1694,6 @@ class Ports(object):
               shutil.copytree(path, target)
               Ports.clear_project_build(name)
             return
-      finally:
-        shared.Cache.release_cache_lock()
 
     if is_tarbz2:
       fullpath = fullname + '.tar.bz2'
@@ -1791,10 +1707,6 @@ class Ports(object):
       logger.debug('    (at ' + fullname + ')')
       Ports.name_cache.add(name)
 
-    class State(object):
-      retrieved = False
-      unpacked = False
-
     def retrieve():
       # retrieve from remote server
       logger.info('retrieving port: ' + name + ' from ' + url)
@@ -1803,15 +1715,9 @@ class Ports(object):
         response = requests.get(url)
         data = response.content
       except ImportError:
-        try:
-          from urllib.request import urlopen
-          f = urlopen(url)
-          data = f.read()
-        except ImportError:
-          # Python 2 compatibility
-          from urllib2 import urlopen
-          f = urlopen(url)
-          data = f.read()
+        from urllib.request import urlopen
+        f = urlopen(url)
+        data = f.read()
 
       if sha512hash:
         actual_hash = hashlib.sha512(data).hexdigest()
@@ -1819,7 +1725,6 @@ class Ports(object):
           shared.exit_with_error('Unexpected hash: ' + actual_hash + '\n'
                                  'If you are updating the port, please update the hash in the port module.')
       open(fullpath, 'wb').write(data)
-      State.retrieved = True
 
     def check_tag():
       if is_tarbz2:
@@ -1836,49 +1741,30 @@ class Ports(object):
     def unpack():
       logger.info('unpacking port: ' + name)
       shared.safe_ensure_dirs(fullname)
+      shutil.unpack_archive(filename=fullpath, extract_dir=fullname)
 
-      # TODO: Someday when we are using Python 3, we might want to change the
-      # code below to use shlib.unpack_archive
-      # e.g.: shutil.unpack_archive(filename=fullpath, extract_dir=fullname)
-      # (https://docs.python.org/3/library/shutil.html#shutil.unpack_archive)
-      if is_tarbz2:
-        z = tarfile.open(fullpath, 'r:bz2')
-      elif url.endswith('.tar.gz'):
-        z = tarfile.open(fullpath, 'r:gz')
-      else:
-        z = zipfile.ZipFile(fullpath, 'r')
-      try:
-        cwd = os.getcwd()
-        os.chdir(fullname)
-        z.extractall()
-      finally:
-        os.chdir(cwd)
-
-      State.unpacked = True
+    # before acquiring the lock we have an early out if the port already exists
+    if os.path.exists(fullpath) and check_tag():
+      return
 
     # main logic. do this under a cache lock, since we don't want multiple jobs to
     # retrieve the same port at once
-
-    shared.Cache.acquire_cache_lock()
-    try:
-      if not os.path.exists(fullpath):
-        retrieve()
-
-      if not os.path.exists(fullname):
-        unpack()
-
-      if not check_tag():
+    with shared.Cache.lock():
+      if os.path.exists(fullpath):
+        # Another early out in case another process build the library while we were
+        # waiting for the lock
+        if check_tag():
+          return
+        # file exists but tag is bad
         logger.warning('local copy of port is not correct, retrieving from remote server')
         shared.try_delete(fullname)
         shared.try_delete(fullpath)
-        retrieve()
-        unpack()
 
-      if State.unpacked:
-        # we unpacked a new version, clear the build in the cache
-        Ports.clear_project_build(name)
-    finally:
-      shared.Cache.release_cache_lock()
+      retrieve()
+      unpack()
+
+      # we unpacked a new version, clear the build in the cache
+      Ports.clear_project_build(name)
 
   @staticmethod
   def clear_project_build(name):
@@ -1959,7 +1845,7 @@ def build_port(port_name, settings):
 def add_ports_cflags(args, settings):
   # Legacy SDL1 port is not actually a port at all but builtin
   if settings.USE_SDL == 1:
-    args += ['-Xclang', '-isystem' + shared.path_from_root('system', 'include', 'SDL')]
+    args += ['-Xclang', '-iwithsysroot/include/SDL']
 
   needed = get_needed_ports(settings)
 
@@ -1976,3 +1862,48 @@ def show_ports():
   print('Available ports:')
   for port in ports.ports:
     print('   ', port.show())
+
+
+# Once we require python 3.8 we can use shutil.copytree with
+# dirs_exist_ok=True and remove this function.
+def copytree_exist_ok(src, dest):
+  with utils.chdir(src):
+    for dirname, dirs, files in os.walk('.'):
+      destdir = os.path.join(dest, dirname)
+      utils.safe_ensure_dirs(destdir)
+      for f in files:
+        shared.safe_copy(os.path.join(src, dirname, f), os.path.join(destdir, f))
+
+
+def install_system_headers(stamp):
+  install_dirs = {
+    ('include',): '',
+    ('lib', 'compiler-rt', 'include'): '',
+    ('lib', 'libunwind', 'include'): '',
+    # Copy the generic arch files first then
+    ('lib', 'libc', 'musl', 'arch', 'generic'): '',
+    # Then overlay the emscripten directory on top.
+    # This mimicks how musl itself installs its headers.
+    ('lib', 'libc', 'musl', 'arch', 'emscripten'): '',
+    ('lib', 'libc', 'musl', 'include'): '',
+    ('lib', 'libcxx', 'include'): os.path.join('c++', 'v1'),
+    ('lib', 'libcxxabi', 'include'): os.path.join('c++', 'v1'),
+  }
+
+  target_include_dir = shared.Cache.get_include_dir()
+  for src, dest in install_dirs.items():
+    src = shared.path_from_root('system', *src)
+    dest = os.path.join(target_include_dir, dest)
+    copytree_exist_ok(src, dest)
+
+  # Create a stamp file that signal the the header have been installed
+  # Removing this file, or running `emcc --clear-cache` or running
+  # `./embuilder build sysroot --force` will cause the re-installation of
+  # the system headers.
+  with open(stamp, 'w') as f:
+    f.write('x')
+  return stamp
+
+
+def ensure_sysroot():
+  shared.Cache.get('sysroot_install.stamp', install_system_headers, what='system headers')
